/*
 * =====================================================================
 * OpenStreetMapSAXHandler.java: SAX parser for open street map files...
 * 
 * Written by: Mark Austin                                  October 2015 
 * =====================================================================
 */

package osm01;

import java.io.File;
import java.lang.Math;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.*;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import osm01.model.*;

public class OpenStreetMapSAXHandler extends DefaultHandler {
   boolean bFirstName = false;
   boolean bLastName  = false;
   boolean bNickName  = false;
   boolean bMarks     = false;

   // element and object stacks ...

   private final Stack<String> elementStack = new Stack<String>();
   private final Stack<Object> objectStack  = new Stack<Object>();

   // memory for nameset and working element stack ...

   private final SortedSet<String>     nodeNameSet = new TreeSet<String>();
   private final SortedSet<String>      wayNameSet = new TreeSet<String>();
   private final SortedSet<String> relationNameSet = new TreeSet<String>();

   // allocate sets for basic / exhaustive storage ....

   private Bounds                bounds = new Bounds();
   private Set<OpenStreetMapNode> nodes = new HashSet<OpenStreetMapNode>();
   private Set<Way>                ways = new HashSet<Way>();
   private Set<Relation>      relations = new HashSet<Relation>();

   // Hashmap(<String>, Set <Way>) various types of ways ...

   private TreeMap<String, OpenStreetMapNode> osmNodeMap = new TreeMap <String, OpenStreetMapNode> ();
   private TreeMap<String, Way>                osmWayMap = new TreeMap <String, Way> ();
   private TreeMap<String, Relation>      osmRelationMap = new TreeMap <String, Relation> ();
   private TreeMap<String, HashSet<OpenStreetMapNode>> osmNodeTypeMap = new TreeMap <String, HashSet<OpenStreetMapNode>> ();
   private TreeMap<String, HashSet<Way>>      osmWayTypeMap      = new TreeMap <String, HashSet<Way>> ();
   private TreeMap<String, HashSet<Relation>> osmRelationTypeMap = new TreeMap <String, HashSet<Relation>> ();

   boolean DEBUG = false;

   @Override
   public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {

      if( DEBUG == true ) {
         System.out.printf("*** In UserHandler.startElement() : qName = %s ... \n", qName );
      }

      // ===================================================================
      // Part 1: Process attributes of nodes, ways and relation tags ...
      // ===================================================================

      // ----------------------------------------
      // Part 1: process attributes of nodes ....
      // ----------------------------------------

      if ("node".equals(qName) ) {
         int length = attributes.getLength();

         if( DEBUG == true ) {
            System.out.printf("--- \n" );
            System.out.printf("--- Start node ... \n" );
            System.out.printf("--- No attributes = %d ... \n", length );
         }

         String changeset = null;
         String        id = null;
         String       lat = null;
         String       lon = null;
         String timestamp = null;
         String uid       = null;
         String user      = null;
         String version   = null;
         String visible   = null;

         for (int i=0; i<length; i++) {
             String  name = attributes.getQName(i);
             String value = attributes.getValue(i);

             if( DEBUG == true ) {
                 System.out.printf("--- %4d name = %s value = %s ... \n", i, name, value );
             }

             if ("id".equals(name) ) {
                id = attributes.getValue(i);
             }
             if ("lat".equals(name) ) {
                lat = attributes.getValue(i);
             }
             if ("lon".equals(name) ) {
                lon = attributes.getValue(i);
             }
             if ("uid".equals(name) ) {
                uid = attributes.getValue(i);
             }
             if ("user".equals(name) ) {
                user = attributes.getValue(i);
             }
         }

         // Store details of new node ...

         if ( id != null && lon != null && lat != null ) {

            System.out.printf("--- NODE: id = %s [lon,lat] = [%s, %s] ... \n", id, lon, lat );

            // Create OpenStreetMapNode data structure ...
            // Add OpenStreetMapNode to TreeMap data structure ...

            OpenStreetMapNode osmNode = new OpenStreetMapNode(id, visible, timestamp,
                                        version, changeset, user, uid, lat, lon, null );
            osmNodeMap.put (id, osmNode );

            // Push osmNode onto the object stack ...

            objectStack.push ( osmNode );
         }
      }

      // ----------------------------------------
      // Part 1: process attributes of ways ....
      // ----------------------------------------

      if ("way".equals(qName) ) {
         int length = attributes.getLength();

         if( DEBUG == true ) {
            System.out.printf("--- \n" );
            System.out.printf("--- Start way ... \n" );
            System.out.printf("--- No attributes = %d ... \n", length );
         }

         String id = null;
         String  v = null;
         String  k = null;
         String visible   = null;
         String timestamp = null;
         String version   = null;
         String changeset = null;
         String uid  = null;
         String user = null;
         for (int i=0; i<length; i++) {
             String  name = attributes.getQName(i);
             String value = attributes.getValue(i);

             if( DEBUG == true ) {
                System.out.printf("--- %4d name = %s value = %s ... \n", i, name, value );
             }

             if ("id".equals(name) ) {
                id = attributes.getValue(i);
             }
         }

         // Print summary of way attributes ....

         if (DEBUG == true) {
            System.out.printf("--- WAY: ");
            if ( id != null ) System.out.printf(" id = %s ", id );
            System.out.printf(" ... \n" );
         }

         // Create Way object and add to TreeMap data structure ...

         Way osmWay = new Way( id, visible, timestamp, version,
                      changeset, user, uid, null, null );
         osmWayMap.put (id, osmWay );

         objectStack.push ( osmWay );
      }

      // --------------------------------------------
      // Part 1: process attributes of relations ....
      // --------------------------------------------

      if ("relation".equals(qName) ) {
         int length = attributes.getLength();

         if (DEBUG == true) {
            System.out.printf("--- \n" );
            System.out.printf("--- Start relation ... \n" );
            System.out.printf("--- No attributes = %d ... \n", length );
         }

         String changeset = null;
         String        id = null;
         String timestamp = null;
         String uid       = null;
         String user      = null;
         String version   = null;
         String visible   = null;
         for (int i=0; i<length; i++) {
             String  name = attributes.getQName(i);
             String value = attributes.getValue(i);

             if (DEBUG == true) {
                System.out.printf("--- %4d name = %s value = %s ... \n", i, name, value );
             }

             if ("id".equals(name) ) {
                id = attributes.getValue(i);
             }
         }

         // Create Relation object and add to TreeMap data structure ...

         Relation osmRelation = new Relation(id, visible, timestamp, version,
                                changeset, user, uid, null, null );
         osmRelationMap.put (id, osmRelation );

         objectStack.push ( osmRelation );
      }

      // ===================================================================
      // Part 2: Process tag attributes inside nodes, ways and relations ...
      // ===================================================================

      // ---------------------------------------------------
      // Part 2 : process tags within nodes ....
      // ---------------------------------------------------

      if ("tag".equals(qName) && "node".equals( elementStack.peek() )) {

         // Retrieve current working OSM node ...

         OpenStreetMapNode osmNode = (OpenStreetMapNode) this.objectStack.peek();

         // Get tag attributes ...

         String  name01 = attributes.getQName(0);
         String value01 = attributes.getValue(0);
         String  name02 = attributes.getQName(1);
         String value02 = attributes.getValue(1);

         // Deal with tags ....

         if ("amenity".equals(value01) == true ||
         "addr:street".equals(value01) == true ||
           "addr:city".equals(value01) == true ||
             "aeroway".equals(value01) == true ||
             "barrier".equals(value01) == true ||
             "bicycle".equals(value01) == true ||
              "bridge".equals(value01) == true ||
            "building".equals(value01) == true ||
             "highway".equals(value01) == true ||
             "landuse".equals(value01) == true ||
                "name".equals(value01) == true ||
              "oneway".equals(value01) == true ||
             "parking".equals(value01) == true ||
             "railway".equals(value01) == true ||
                "shop".equals(value01) == true ||
              "tunnel".equals(value01) == true ||
            "waterway".equals(value01) == true ) {

             if (DEBUG == true) {
                System.out.printf("---    TAG: %s = %s ... \n", value01, value02 );
             }
             osmNode.addTag ( value01, value02 );
         }

         // Add tags of type "name" to nodeNameSet ... 

         if ("name".equals(value01) == true) {

            if (DEBUG == true) {
               System.out.printf("---    Adding %s to nodeNameSet () ... \n", value02 );
            }

            nodeNameSet.add( value02 );
         }
      }

      // -----------------------------------------------------------
      // Part 2: process "nd" elements embedded within way tags ....
      // -----------------------------------------------------------

      if ("nd".equals(qName) && "way".equals( elementStack.peek() )) {

         Way osmWay = (Way) this.objectStack.peek();

         // get ref and string attributes ...

         String  name = attributes.getQName(0);
         String value = attributes.getValue(0);
         if ( name != null && DEBUG == true )
            System.out.printf("---    ND: ref = %s ... \n", value );

         // Add "nd" reference node to way element ....

         OpenStreetMapNode osmNode = (OpenStreetMapNode) osmNodeMap.get ( value );
         if ( osmNode == null ) {
            System.out.printf("--- ERROR ND: ref = %s not found ... \n", value );
         } else {
            osmWay.addNode ( osmNode );
         }
      }

      // -------------------------------------------------
      // Part 2: process "tag" elements embedded within ways ....
      // -------------------------------------------------

      if ("tag".equals(qName) && "way".equals( elementStack.peek() )) {

         Way osmWay = (Way) this.objectStack.peek();

         int length = attributes.getLength();
         if( DEBUG == true ) {
            System.out.printf("--- Way: no attributes = %d ... \n", length );
         }

         // Get tag attributes ...

         String  name01 = attributes.getQName(0);
         String value01 = attributes.getValue(0);
         String  name02 = attributes.getQName(1);
         String value02 = attributes.getValue(1);

         // Deal with tags ....

         if ("amenity".equals(value01) == true ||
         "addr:street".equals(value01) == true ||
           "addr:city".equals(value01) == true ||
             "aeroway".equals(value01) == true ||
             "barrier".equals(value01) == true ||
             "bicycle".equals(value01) == true ||
              "bridge".equals(value01) == true ||
            "building".equals(value01) == true ||
             "highway".equals(value01) == true ||
             "landuse".equals(value01) == true ||
                "name".equals(value01) == true ||
              "oneway".equals(value01) == true ||
             "parking".equals(value01) == true ||
             "railway".equals(value01) == true ||
                "shop".equals(value01) == true ||
              "tunnel".equals(value01) == true ||
            "waterway".equals(value01) == true ) {

             if (DEBUG == true) {
                System.out.printf("---    TAG: %s = %s ... \n", value01, value02 );
             }
             osmWay.addTag ( value01, value02 );
         }
      }

      // -------------------------------------------------
      // Part 2: process relation tags and attributes ....
      // -------------------------------------------------

      if ("tag".equals(qName) && "relation".equals( elementStack.peek() )) {

         Relation osmRelation = (Relation) this.objectStack.peek();

         // Get tag attributes ...

         String  name01 = attributes.getQName(0);
         String value01 = attributes.getValue(0);
         String  name02 = attributes.getQName(1);
         String value02 = attributes.getValue(1);

         if ("amenity".equals(value01) == true ||
         "addr:street".equals(value01) == true ||
           "addr:city".equals(value01) == true ||
             "aeroway".equals(value01) == true ||
             "barrier".equals(value01) == true ||
             "bicycle".equals(value01) == true ||
              "bridge".equals(value01) == true ||
            "building".equals(value01) == true ||
             "highway".equals(value01) == true ||
             "landuse".equals(value01) == true ||
                "name".equals(value01) == true ||
              "oneway".equals(value01) == true ||
             "parking".equals(value01) == true ||
             "railway".equals(value01) == true ||
                "shop".equals(value01) == true ||
              "tunnel".equals(value01) == true ||
            "waterway".equals(value01) == true ) {

             if (DEBUG == true) {
                System.out.printf("---    TAG: %s = %s ... \n", value01, value02 );
             }

             osmRelation.addTag ( value01, value02 );
         }
      }

      // -----------------------------------------
      // Process relation members and attributes ....
      // -----------------------------------------

      if ("member".equals(qName) && "relation".equals( elementStack.peek() )) {

         Relation osmRelation = (Relation) this.objectStack.peek();

         int length = attributes.getLength();
         if (DEBUG == true) {
            System.out.printf("--- Member: no attributes = %d ... \n", length );
         }

         String  name01 = attributes.getQName(0);
         String value01 = attributes.getValue(0);
         String  name02 = attributes.getQName(1);
         String value02 = attributes.getValue(1);
         String  name03 = attributes.getQName(2);
         String value03 = attributes.getValue(2);

         // Check to see if the way item is actually defined ...

         if ("way".equals(value01) == true && DEBUG == true ) {
             Way w = (Way) osmWayMap.get ( value02 );
             if ( w == null )
                System.out.printf("--- RELATION ERROR: Way[ %s ] not defined ... \n", value02 );
             else
                System.out.printf("--- RELATION CORRECT: Way[ %s ] found ... \n", value02 );
         }

         // Add new member to relation object ...

         Member m = new Member ( value01, value02, value03 );
         osmRelation.addMember( m );
      }

      if (DEBUG == true) {
         System.out.printf("*** In startElement(): elementStack.push(qName) ... \n");
      }

      elementStack.push(qName);
   }

   @Override
   public void endElement(String uri, String localName, String qName) throws SAXException {

      if (DEBUG == true) {
         System.out.printf("*** In endElement(): qName = %s ... \n", qName );
      }

      // Pop name off element stack ...

      elementStack.pop();

      // Pop nodes, ways and relations from object stack ...

      if ( "node".equals(qName) == true || "way".equals(qName)  == true ||
           "relation".equals(qName) == true ) {

           if (DEBUG == true) 
              System.out.printf("*** In endElement(): popping object stack ... \n");

           objectStack.pop();
      }
   }

   @Override
   public void characters(char ch[], 
      int start, int length) throws SAXException {
      if (bFirstName) {
         System.out.println("First Name: " + new String(ch, start, length));
         bFirstName = false;
      } else if (bLastName) {
         System.out.println("Last Name: " + new String(ch, start, length));
         bLastName = false;
      } else if (bNickName) {
         System.out.println("Nick Name: " + new String(ch, start, length));
         bNickName = false;
      } else if (bMarks) {
         System.out.println("Marks: " + new String(ch, start, length));
         bMarks = false;
      }
   }

   // ==========================================================
   // Retrieve parsed data and references to HashMaps ....
   // ==========================================================
    
   public SortedSet<String> getNodeNames() {
      return nodeNameSet;
   }
    
   public SortedSet<String> getWayNames() {
      return wayNameSet;
   }
    
   public SortedSet<String> getRelationNames() {
      return relationNameSet;
   }

   public TreeMap<String, OpenStreetMapNode> getOpenStreetMapNodeTreeMap() {
      return osmNodeMap;
   }

   public TreeMap<String, Way> getOpenStreetMapWayTreeMap() {
      return osmWayMap;
   }

   public TreeMap<String, Relation> getOpenStreetMapRelationTreeMap() {
      return osmRelationMap;
   }
}
