/**
 * =====================================================
 * AbstractNode.java: .....
 *
 * Written by: Mark Austin and Willy Tiengo
 * =====================================================
 */

package osm01.model;

import java.util.Map;
import java.util.HashMap;

public class AbstractNode {
    public String id;
    public String visible;
    public String timestamp;
    public String version;
    public String changeset;
    public String user;
    public String uid;
    public Map<String, String> tags;

    // --- Keywords for OSM node, way and relation types ---

    public static final String AEROWAY  = "aeroway";
    public static final String AMENITY  = "amenity";
    public static final String APRON    = "apron";
    public static final String BICYCLE  = "bicycle";
    public static final String BOUNDARY = "boundary";
    public static final String BUILDING = "building";
    public static final String COASTLINE = "coastline";
    public static final String CROSSING = "crossing";
    public static final String ENTRANCE = "entrance";
    public static final String FOOTWAY  = "footway";
    public static final String GATE     = "gate";
    public static final String HIGHWAY  = "highway";
    public static final String HOLDING  = "holding_position";
    public static final String JUNCTION = "junction";
    public static final String LEISURE  = "leisure";
    public static final String LIGHTRAIL    = "light_rail";
    public static final String MOTORWAY     = "motorway";
    public static final String MOTORWAYLINK = "motorway_link";
    public static final String MULTIPOLYGON = "multipolygon";
    public static final String NAME     = "name";
    public static final String NATURAL  = "natural";
    public static final String NETWORK  = "network";
    public static final String PLACE    = "place";
    public static final String PRIMARY  = "primary";
    public static final String RAILWAY  = "railway";
    public static final String RESIDENTIAL = "residential";
    public static final String ROUTE    = "route";
    public static final String RUNWAY   = "runway";
    public static final String SECONDARY = "secondary";
    public static final String SERVICE  = "service";
    public static final String SHOP     = "shop";
    public static final String SITE     = "site";
    public static final String STATION  = "station";
    public static final String SUBWAY   = "subway";
    public static final String TAXIWAY  = "taxiway";
    public static final String TERMINAL = "terminal";
    public static final String TOURISM  = "tourism";
    public static final String TRACK    = "track";
    public static final String TUNNEL   = "tunnel";

    // --- Collections of Networks / Routes ---

    public static final String ROUTEMASTER = "route_master";

    // --- Leonard's keywords ---

    public static final String AWAY       = "away";
    public static final String CLOSE      = "close";
    public static final String LINE       = "line";
    public static final String LINEBLOCK  = "lineblock";
    public static final String SPACEBLOCK = "spaceblock";
    public static final String INTERSECTION = "intersection";

    // Constructor methods ....

    public AbstractNode() {}

    public AbstractNode(String id, String visible, String timestamp,
            String version, String changeset, String user, String uid,
            Map<String, String> tags) {

        this.id        = id;
        this.visible   = visible;
        this.timestamp = timestamp;
        this.version   = version;
        this.changeset = changeset;
        this.user      = user;
        this.uid       = uid;

        if (tags == null)
           this.tags = new HashMap();
        else
           this.tags = tags;
    }

    // --------------------------------------------
    // Add item to set of tags .....
    // --------------------------------------------

    public void addTag ( String key, String value ) {
       tags.put (key, value);
    }

    // --------------------------------------------
    // Retrieve details of the Abstract Node .....
    // --------------------------------------------

    public String getId() {
       return id;
    }

    public Map<String, String> getTags() {
       return tags;
    }

    // ----------------------------------------------------------
    // Retrieve details of node and way tags ....
    // ----------------------------------------------------------

    public boolean hasName() {
        return (tags.get("name") != null);
    }

    public String getName() {
        return tags.get("name");
    }

    // ----------------------------------------------------------
    // Methods for Leonard's keywords ....
    // ----------------------------------------------------------

    // Traffic Intersection ... 

    public boolean isTrafficIntersection() {
        String s = (String) tags.get( SPACEBLOCK );
        if (s != null && s.equals( INTERSECTION ) == true )
           return true;
        else 
           return false;
    }

    public String getTrafficIntersectionType() {
        String s = (String) tags.get( NAME );
        if (s != null)
           return s;
        else
           return tags.get( SPACEBLOCK );
    }

    // Regular lineblock ... 

    public boolean isLineBlock() {
        String s = (String) tags.get( SPACEBLOCK );
        if (s != null && s.equals( LINE ) == true )
           return true;
        else 
           return false;
    }

    public String getLineBlockType() {
        String s = (String) tags.get( NAME );
        if (s != null)
           return s;
        else
           return tags.get( SPACEBLOCK );
    }

    // Regular spaceblock ... 

    public boolean isSpaceBlock() {
        String s = (String) tags.get( SPACEBLOCK );
        boolean blockType = false;
        if (s != null) {
           blockType = (s.equals( CLOSE ) == true ||
                        s.equals( INTERSECTION ) == true ||
                        s.equals( AWAY ) == true );
        }
        if (blockType == true )
           return true;
        else 
           return false;
    }

    public String getSpaceBlockType() {
        String s = (String) tags.get( NAME );
        if (s != null)
           return s;
        else
           return tags.get( SPACEBLOCK );
    }

    // ----------------------------------------------------------
    // Collections of Networks / Routes ....
    // ----------------------------------------------------------

    public boolean isRouteMaster() {
        String s = (String) tags.get("type");
        if (s != null && s.equals( ROUTEMASTER ) == true )
           return true;
        else 
           return false;
    }

    public String getRouteMasterType() {
        return tags.get( ROUTEMASTER );
    }

    // ----------------------------------------------------------
    // Retrieve details of node and way tags ....
    // ----------------------------------------------------------

    // Aiport Apron ... 

    public boolean isAirportApron() {
        String s = (String) tags.get(AEROWAY);
        if (s != null && s.equals( APRON ) == true )
           return true;
        else 
           return false;
    }

    public String getAirportApron() {
        return tags.get(AEROWAY);
    }

    // Amenity ... 

    public boolean isAmenity() {
        return (tags.get(AMENITY) != null);
    }

    public String getAmenityType() {
        return tags.get(AMENITY);
    }

    // Bicycle ... 

    public boolean isBicycle() {
        return (tags.get(BICYCLE) != null);
    }

    public String getBicycleType() {
        return tags.get(BICYCLE);
    }

    // Boundary ... 

    public boolean isBoundary() {
        return (tags.get(BOUNDARY) != null);
    }

    public String getBoundaryType() {
        return tags.get(BOUNDARY);
    }

    // Building ... 

    public boolean isBuilding() {
        return (tags.get(BUILDING) != null);
    }

    public String getBuildingType() {
        return tags.get(BUILDING);
    }

    // Natural coastline ... 

    public boolean isCoastline() {
        String s = (String) tags.get(NATURAL);
        if (s != null && s.equals( COASTLINE ) == true)
           return true;
        else 
           return false;
    }

    public String getCoastlineType() {
        return tags.get(NATURAL);
    }

    // Crossing ... 

    public boolean isCrossing() {
        return (tags.get(CROSSING) != null);
    }

    public String getCrossingType() {
        return tags.get(CROSSING);
    }

    // Entrance ... 

    public boolean isEntrance() {
        return (tags.get(ENTRANCE) != null);
    }

    public String getEntranceType() {
        return tags.get(ENTRANCE);
    }

    // Footway ... 

    public boolean isFootway() {
        String s = (String) tags.get(HIGHWAY);
        if (s != null && s.equals( FOOTWAY ) == true)
           return true;
        else 
           return false;
    }

    public String getFootwayType() {
        return tags.get(HIGHWAY);
    }

    // Main highway ... 

    public boolean isHighway() {
        String s = (String) tags.get(HIGHWAY);
        boolean highwayType = false;
        if (s != null) {
           highwayType = (s.equals( MOTORWAY ) == true || s.equals( MOTORWAYLINK ) == true ||
                           s.equals( PRIMARY ) == true || s.equals( SECONDARY ) == true );
        }

        if (highwayType == true )
           return true;
        else 
           return false;
    }

    public String getHighwayType() {
        return tags.get(HIGHWAY);
    }

    // Junction ... 

    public boolean isJunction() {
        return (tags.get(JUNCTION) != null);
    }

    public String getJunctionType() {
        return tags.get(JUNCTION);
    }

    // Leisure ...

    public boolean isLeisure() {
        return (tags.get(LEISURE) != null);
    }

    public String getLeisureType() {
        return tags.get(LEISURE);
    }

    // Multipolygon ...

    public boolean isMultiPolygon() {
        String s = (String) tags.get("type");
        if (s != null && s.equals( MULTIPOLYGON ) == true )
           return true;
        else 
           return false;
    }

    public String getMultiPolygonType() {
        return tags.get(MULTIPOLYGON);
    }

    // Natural ...

    public boolean isNatural() {
        return (tags.get(NATURAL) != null);
    }

    public String getNaturalType() {
        return tags.get(NATURAL);
    }

    // Network ...

    public boolean isNetwork() {
        return (tags.get(NETWORK) != null);
    }

    public String getNetworkType() {
        return tags.get(NETWORK);
    }

    // Place ...

    public boolean isPlace() {
        return (tags.get(PLACE) != null);
    }

    public String getPlaceType() {
        return tags.get(PLACE);
    }

    // Railway ... 

    public boolean isRailway() {
        String s = (String) tags.get(RAILWAY);
        if (s != null && s.equals( LIGHTRAIL ) == true )
           return true;
        else 
           return false;
    }

    public String getRailwayType() {
        return tags.get(RAILWAY);
    }

    // Railway station ... 

    public boolean isRailwayStation() {
        String s = (String) tags.get(RAILWAY);
        if (s != null && s.equals( STATION ) == true )
           return true;
        else 
           return false;
    }

    public String getRailwayStationType() {
        return tags.get(RAILWAY);
    }

    // Residential streets ... 

    public boolean isResidential() {
        String s = (String) tags.get(HIGHWAY);
        if (s != null && s.equals( RESIDENTIAL ) == true )
           return true;
        else 
           return false;
    }

    public String getResidentialType() {
        return tags.get(HIGHWAY);
    }

    // Route ... 

    public boolean isRoute() {
        return (tags.get(ROUTE) != null);
    }

    public String getRouteType() {
        return tags.get(ROUTE);
    }

    // Runway ... 

    public boolean isRunway() {
        String s = (String) tags.get(AEROWAY);
        if (s != null && s.equals( RUNWAY ) == true )
           return true;
        else 
           return false;
    }

    public String getRunwayType() {
        return tags.get(AEROWAY);
    }

    // Service roads ... 

    public boolean isServiceRoad() {
        String s = (String) tags.get(HIGHWAY);
        boolean highwayType = false;

        if (s != null) {
           highwayType = ( s.equals( SERVICE ) == true );
        }

        if (highwayType == true )
           return true;
        else 
           return false;
    }

    public String getServiceRoadType() {
        return tags.get(HIGHWAY);
    }

    // Shop ... 

    public boolean isShop() {
        return (tags.get(SHOP) != null);
    }

    public String getShopType() {
        return tags.get(SHOP);
    }

    // Site ... 

    public boolean isSite() {
        return (tags.get(SITE) != null);
    }

    public String getSiteType() {
        return tags.get(SITE);
    }

    // Subway ... 

    public boolean isSubway() {
        String s = (String) tags.get(RAILWAY);
        if (s != null && s.equals( SUBWAY ) == true )
           return true;
        else 
           return false;
    }

    public String getSubwayType() {
        return tags.get(RAILWAY);
    }

    // Taxiway ... 

    public boolean isTaxiway() {
        String s = (String) tags.get(AEROWAY);
        if (s != null && s.equals( TAXIWAY ) == true )
           return true;
        else 
           return false;
    }

    public String getTaxiwayType() {
        return tags.get(AEROWAY);
    }

    // Taxiway Holding Position ... 

    public boolean isTaxiwayHoldingPosition() {
        String s = (String) tags.get(AEROWAY);
        if (s != null && s.equals( HOLDING ) == true )
           return true;
        else 
           return false;
    }

    public String getTaxiwayHoldingPositionType() {
        return tags.get(AEROWAY);
    }


    // Aiport Terminal Building ... 

    public boolean isAirportTerminal() {
        String s = (String) tags.get(AEROWAY);
        if (s != null && s.equals( TERMINAL ) == true )
           return true;
        else 
           return false;
    }

    public String getAirportTerminal() {
        return tags.get(AEROWAY);
    }

    // Tourism ... 

    public boolean isTourism() {
        return (tags.get(TOURISM) != null);
    }

    public String getTourismType() {
        return tags.get(TOURISM);
    }

    // Tunnel ... 

    public boolean isTunnel() {
        return (tags.get(TUNNEL) != null);
    }

    public String getTunnelType() {
        return tags.get(TUNNEL);
    }


    // --------------------------------------------
    // Hashmap methods .....
    // --------------------------------------------

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }

        final OpenStreetMapNode other = (OpenStreetMapNode) obj;

        return (id != null ? id.equals(other.id) : false);
    }

    @Override
    public int hashCode() {
        return (id != null ? id.hashCode() : super.hashCode());
    }
}
