/*
 * =====================================================================
 * OpenStreetMapModel.java: Open Street Map Model that interacts with a 
 * SAX parser .... 
 * 
 * Written by: Mark Austin                                  October 2015 
 * =====================================================================
 */

package osm02.model;

import java.io.*;
import java.io.File;
import java.lang.Math;
import java.net.URL;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.*;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.*;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.*;

// import whistle.*;
// import whistle.model.*;
// import whistle.model.primitive.Circle;
// import whistle.model.primitive.Point;
// import whistle.model.primitive.Edge;
// import whistle.util.osm.model.*;
// import whistle.util.osm.parser.*;

import osm02.parser.*;

public class OpenStreetMapModel {

   // element and object stacks ...

   private final Stack<String> elementStack;
   private final Stack<Object> objectStack;

   // memory for nameset and working element stack ...

   private final SortedSet<String>     nodeNameSet;
   private final SortedSet<String>      wayNameSet;
   private final SortedSet<String> relationNameSet;

   // allocate sets for basic / exhaustive storage ....

   private Bounds                bounds;
   private Set<OpenStreetMapNode> nodes;
   private Set<Way>                ways;
   private Set<Relation>      relations;

   // Hashmap(<String>, Set <Way>) various types of ways ...

   private TreeMap<String, OpenStreetMapNode> osmNodeMap;
   private TreeMap<String, Way>                osmWayMap;
   private TreeMap<String, Relation>      osmRelationMap;
   private TreeMap<String, HashSet<OpenStreetMapNode>> osmNodeTypeMap;
   private TreeMap<String, HashSet<Way>>      osmWayTypeMap;
   private TreeMap<String, HashSet<Relation>> osmRelationTypeMap;

   boolean DEBUG = false;
   private OpenStreetMapSAXHandler osmhandler = null;

   // Dot size for workspace visualization ..

   private int    smallDotSize = 6;
   private double scaleFactor  = 1.0;

   // -------------------------------------------------------
   // Constructor methods ...
   // -------------------------------------------------------

   public OpenStreetMapModel() {

      // element and object stacks ...

      elementStack = new Stack<String>();
      objectStack  = new Stack<Object>();

      // memory for nameset and working element stack ...

      nodeNameSet     = new TreeSet<String>();
      wayNameSet      = new TreeSet<String>();
      relationNameSet = new TreeSet<String>();

      // allocate sets for basic / exhaustive storage ....

      bounds = new Bounds();
      nodes  = new HashSet<OpenStreetMapNode>();
      ways   = new HashSet<Way>();
      relations = new HashSet<Relation>();

      // various types of treemap relationships ...

      osmNodeMap = new TreeMap <String, OpenStreetMapNode> ();
      osmWayMap  = new TreeMap <String, Way> ();
      osmRelationMap = new TreeMap <String, Relation> ();

      osmNodeTypeMap = new TreeMap <String, HashSet<OpenStreetMapNode>> ();
      osmWayTypeMap      = new TreeMap <String, HashSet<Way>> ();
      osmRelationTypeMap = new TreeMap <String, HashSet<Relation>> ();
   }

   // ======================================================
   // Use SAX implementation of OSM Parser to read file ....
   // ======================================================

   public void getData( String osmFileName ) throws Exception {

       if ( DEBUG == true ) {
           System.out.printf("*** In OpenStreetMapModel.getData(): IMPORTING OSM file ... \n");
           System.out.printf("*** In OpenStreetMapModel.getData(): Name = %s ... \n", osmFileName );
       }

       // Remove quotes from the xml filename (if they exist) ...

       osmFileName = osmFileName.replace("\"", "");

       // --------------------------------------------
       // Check that file name has .osm extension ....
       // --------------------------------------------

       if ( osmFileName.contains(".")) {
          String[] parts = osmFileName.split("\\.");
          String extension = (String) parts[ parts.length - 1 ];
          if (DEBUG == true) {
              System.out.printf("--- Filename extension: %s ... \n", extension );
          }

          if (extension.contains("osm") != true) {
             System.out.printf("--- ERROR: OpenStreetMapModel.getData() ... \n" );
             System.out.printf("--- ERROR: Filename extension needs to be \".osm\"! ...\n" );
          }
       } else {
          throw new IllegalArgumentException("String " + osmFileName + " does not contain .");
       }

       // ---------------------------------------------------------------------------
       // Import data model object to xml, then transfer contents current data model.
       // ---------------------------------------------------------------------------

       File inputFile = new File( osmFileName );
       SAXParserFactory factory = SAXParserFactory.newInstance();
       SAXParser saxParser = factory.newSAXParser();

       osmhandler = new OpenStreetMapSAXHandler();
       saxParser.parse( inputFile, osmhandler);

       // =========================================================
       // Print categories of osm nodes and ways ...
       // =========================================================

       // printOSMNodeTypeMap();
       // printOSMWayTypeMap();
       // printOSMRelationTypeMap();

       // ==================================================================
       // Compute (x,y) coordinates for OSM (long,lat) nodes ...
       // ==================================================================

       if ( DEBUG == true ) {
           System.out.printf("*** In OpenStreetMapModel.getData(): LEAVING OSM IMPORT ... \n");
       }
   }

   // ====================================================
   // Retrieve workspace from OSM model ....
   // ====================================================

   // public CompositeHierarchy getWorkspace( String name ) {
   //        ... fill in details later ....
   // }

    // =======================================================
    // Print names of OSM node, way and relation elements ... 
    // =======================================================

    public void print () {
       SortedSet<String> names = osmhandler.getNodeNames();
       printNames("List of Node Names", names );

       // Print names of OSM Way elements ... 

       names = osmhandler.getWayNames();
       printNames("List of Way Names", names );

       // Print names of OSM Relation elements ... 

       names = osmhandler.getRelationNames();
       printNames("List of Relation Names", names );

       // Print size of OpenStreetMapNode TreeMap ....

       TreeMap<String, OpenStreetMapNode> nodes = osmhandler.getOpenStreetMapNodeTreeMap();
       TreeMap<String, Way>       ways = osmhandler.getOpenStreetMapWayTreeMap();
       TreeMap<String, Relation>  rels = osmhandler.getOpenStreetMapRelationTreeMap();

       System.out.printf("*** No nodes     = %d ... \n", nodes.size() );
       System.out.printf("*** No ways      = %d ... \n", ways.size() );
       System.out.printf("*** No relations = %d ... \n", rels.size() );

       // printOSMNodeTypeMap( nodes );
       // printOSMWayTypeMap( ways );
       // printOSMRelationTypeMap( rels, ways );
   }

   // ===========================================================
   // Print node, way and relation names ...
   // ===========================================================

   public void printNames(String title, SortedSet<String> nameSet ) {
      System.out.printf("*** %s ... \n", title );
      System.out.printf("*** ===============================\n");

      int i = 1;
      for (String name : nameSet) {
          System.out.printf("%4d: %s\n", i, name);
          i = i + 1;
      }
      System.out.println("*** ===============================");
   }

   // =======================================================================
   // Print contents of TreeMap<String, OpenStreetMapNode> osmNodeTypeMap ...
   // =======================================================================

   public void printOSMNodeTypeMap( TreeMap<String, OpenStreetMapNode> osmNodeTypeMap ) {

      // Header for openstreetmap nodes ...

      System.out.println("***");
      System.out.println("*** Summary: TreeMap<String, OpenStreetMapNode> ... ");
      System.out.println("*** ==================================================== ... ");

      // Summary of categories of openstreetmap node ...

      Iterator<Map.Entry<String, OpenStreetMapNode>> iterator = osmNodeTypeMap.entrySet().iterator();
      while(iterator.hasNext()) {
         Map.Entry<String, OpenStreetMapNode> entry = iterator.next();
         String              key = (String) entry.getKey();
         OpenStreetMapNode value = (OpenStreetMapNode) entry.getValue();
         System.out.println( value.toString() );
      }
   }

   // =======================================================================
   // Print contents of TreeMap<String, Way> osmWayTypeMap ...
   // =======================================================================

   public void printOSMWayTypeMap( TreeMap<String, Way> osmWayTypeMap ) {

      // Header for openstreetmap ways ...

      System.out.println("***");
      System.out.println("*** Summary: TreeMap<String, Way> ... ");
      System.out.println("*** ==================================================== ... ");

      // Summary of categories of openstreetmap way ...

      Iterator<Map.Entry<String, Way>> iterator = osmWayTypeMap.entrySet().iterator();
      while(iterator.hasNext()) {
         Map.Entry<String, Way> entry = iterator.next();
         String key = (String) entry.getKey();
         Way  value = (Way) entry.getValue();
         System.out.println( value.toString() );
      }
   }
  
   // =======================================================================
   // Print contents of TreeMap<String, Relation> osmRelationTypeMap ...
   // =======================================================================

   public void printOSMRelationTypeMap( TreeMap<String, Relation> osmRelationTypeMap,
                                        TreeMap<String, Way> osmWayTypeMap ) {

      // Header for openstreetmap ways ...

      System.out.println("***");
      System.out.println("*** Summary: TreeMap<String, Relation> ... ");
      System.out.println("*** ==================================================== ... ");

      // Summary of categories of openstreetmap way ...

      Iterator<Map.Entry<String, Relation>> iterator = osmRelationTypeMap.entrySet().iterator();
      while(iterator.hasNext()) {
         Map.Entry<String, Relation> entry = iterator.next();
         String   key = (String) entry.getKey();
         Relation rel = (Relation) entry.getValue();

         StringBuffer buf = new StringBuffer();
         buf.append( String.format("*** Relation[ name = %s ] ... \n", rel.getName() ) );
         buf.append( String.format("               key = %s ... \n", rel.getId() ) );
         buf.append( String.format("              type = %s ... \n", rel.getType() ) );

         List<Member> members = rel.getMembers();
         if( members != null ) {
            int i = 1;
            for (Member member : members) {
                String wayt = member.type;
                String reft = member.ref;

                buf.append( String.format("*** Member [%3d]: type = %s: ref = %s: role = %s ... \n",
                                            i, member.type, member.ref, member.role ) );

                if( wayt.equals("way") == true ) {
                    Way wayref = (Way) osmWayTypeMap.get (reft);
                    if ( wayref == null ) 
                       buf.append( String.format("***             : ERROR -- way not found ... \n" ));
                    else
                       buf.append( String.format("***             : CORRECT -- way found ... \n" ));
                }

                i = i + 1;
            }
         }

         System.out.println( buf.toString() );
      }
   }
}
