/**
 * =====================================================
 * Way.java: .....
 *
 * @author Willy Tiengo
 * =====================================================
 */

package osm02.model;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.io.WKBWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class Way extends AbstractNode {

    // Attributes ---------------------------------------------------------------

    public List<OpenStreetMapNode> nodes;

    public Way(String id, String visible, String timestamp,
            String version, String changeset, String user,
            String uid, List<OpenStreetMapNode> nodes, Map<String, String> tags) {
        super(id, visible, timestamp, version, changeset, user, uid, tags);

        if ( tags != null )
           this.tags = tags;

        this.nodes = nodes;
    }

    // Add OSM node to Way model ....

    public void addNode ( OpenStreetMapNode nd ) {

       System.out.println("--- In Way.addNode() ... ");

       // Create list for storage of nodes ....

       if ( nodes == null ) {
          nodes = new ArrayList<OpenStreetMapNode>();
       }

       // Add item to list of nodes ....

       nodes.add ( nd );
    }

    // Return arraylist of (lon,lat) nodal ids ....

    public ArrayList getNodalList() {
       ArrayList<OpenStreetMapNode> coords = new ArrayList<OpenStreetMapNode>();
       for (OpenStreetMapNode node : nodes) {
           coords.add ( node );
       }

       return coords;
    }

    // Return arraylist of (lon,lat) nodal coordinates ....

    public ArrayList getNodalArray() {
       ArrayList<Coordinate> coords = new ArrayList<Coordinate>();

       Coordinate c1;
       for (OpenStreetMapNode node : nodes) {
           c1 = new Coordinate( Double.parseDouble(node.lon), Double.parseDouble(node.lat) );
           coords.add ( c1 );
       }

       return coords;
    }

    public LineString getLineString() {
        List<Coordinate> coords = new ArrayList<Coordinate>();
        GeometryFactory fac = new GeometryFactory();

        Coordinate c1;
        for (OpenStreetMapNode node : nodes) {
            c1 = new Coordinate( Double.parseDouble(node.lon), Double.parseDouble(node.lat));
            coords.add(c1);
        }

        return fac.createLineString(coords.toArray(new Coordinate[0]));
    }

    // ----------------------------------------------------------
    // Miscellaneous tag information ....
    // ----------------------------------------------------------

    public boolean isOneway() {
        String oneway = tags.get("oneway");
        return ((oneway != null) ? oneway.equals("yes") : false);
    }

    public double getWayLength() {
        return wayLength(nodes);
    }


    public String getShape() throws Exception {
        MultiLineString mls;

        // Precisa ser um MultiLineString
        mls = new GeometryFactory().createMultiLineString( new LineString[] { getLineString() });

        return WKBWriter.bytesToHex(new WKBWriter().write(mls));
    }

    public String getAltNames() {
        return tags.get("alt_name");
    }

    // ===========================================================================
    // Get coordinate for middle of the list of way points.
    // Used for positioning labels ....
    // ===========================================================================

    public String getWayMiddle() {
        double lenMiddle, distance, lineDistance;
        GeometryFactory fac = new GeometryFactory();

        OpenStreetMapNode n1 = null, n2 = null;

        lenMiddle = wayLength(nodes) / 2;
        distance = 0d;

        for (int i = 0; i < nodes.size() - 1; i++) {

            n1 = nodes.get(i);
            n2 = nodes.get(i + 1);

            lineDistance = lineDistance(n1, n2);

            if ((distance + lineDistance) > lenMiddle) {
                distance = (lenMiddle - distance) / lineDistance;
                break;
            }

            distance += lineDistance;
        }

        double lat = Double.parseDouble(n2.lat);
        double lon = Double.parseDouble(n2.lon);

        if (distance > 0.0d) {
            distance = (1 / distance);

            // Baseado na prova do ponto médio
            lat = (Double.parseDouble(n2.lat) + (distance - 1) * Double.parseDouble(n1.lat)) / distance;
            lon = (Double.parseDouble(n2.lon) + (distance - 1) * Double.parseDouble(n1.lon)) / distance;
        }

        return WKBWriter.bytesToHex(
                new WKBWriter().write(fac.createPoint(new Coordinate(lon, lat))));
    }

    
    // Private methods ---------------------------------------------------------

    private double wayLength( List<OpenStreetMapNode> nodes ) {
        double length = 0d;
        OpenStreetMapNode n1, n2;

        n1 = nodes.get(0);

        for (int i = 1; i < nodes.size(); i++) {
            n2 = nodes.get(i);

            length += LatLongUtil.distance(
                    Double.parseDouble(n1.lat), Double.parseDouble(n1.lon),
                    Double.parseDouble(n2.lat), Double.parseDouble(n2.lon));

            n1 = n2;
        }

        return length;
    }

    private static Double lineDistance(OpenStreetMapNode n1, OpenStreetMapNode n2) {
        return LatLongUtil.distance(
                Double.parseDouble(n1.lat), Double.parseDouble(n1.lon),
                Double.parseDouble(n2.lat), Double.parseDouble(n2.lon));

    }

    // --------------------------------------------------
    // String representation of OpenStreeMap Way ....
    // --------------------------------------------------

    public String toString() {

       StringBuffer buf = new StringBuffer();

       buf.append( String.format(" Way[  key = %10s ] ... \n", getId() ) );

       if (this.hasName() == true)
          buf.append( String.format("      name = %s ... \n",  this.getName() ) );

       if ( this.isAmenity() == true )
          buf.append( String.format("   amenity = %s ... \n", this.getAmenityType() ) );
       if ( this.isBicycle() == true )
          buf.append( String.format("   bicycle = %s ... \n", this.getBicycleType() ) );
       if ( this.isBoundary() == true )
          buf.append( String.format("  boundary = %s ... \n", this.getBoundaryType() ) );
       if ( this.isBuilding() == true )
          buf.append( String.format("  building = %s ... \n", this.getBuildingType() ) );
       if ( this.isCoastline() == true )
          buf.append( String.format(" coastline = %s ... \n", this.getCoastlineType() ) );
       if ( this.isHighway() == true )
          buf.append( String.format("   highway = %s ... \n", this.getHighwayType() ) );
       if ( this.isJunction() == true )
          buf.append( String.format("  junction = %s ... \n", this.getJunctionType() ) );
       if ( this.isLeisure() == true )
          buf.append( String.format("   leisure = %s ... \n", this.getLeisureType() ) );
       if ( this.isNatural() == true )
          buf.append( String.format("   natural = %s ... \n", this.getNaturalType() ) );
       if ( this.isRailway() == true )
          buf.append( String.format("   railway = %s ... \n", this.getRailwayType() ) );
       if ( this.isRoute() == true )
          buf.append( String.format("     route = %s ... \n", this.getRouteType() ) );
       if ( this.isTaxiway() == true )
          buf.append( String.format("   taxiway = %s ... \n", this.getTaxiwayType() ) );
       if ( this.isTourism() == true )
          buf.append( String.format("   tourism = %s ... \n", this.getTourismType() ) );
       if ( this.isTunnel() == true )
          buf.append( String.format("    tunnel = %s ... \n", this.getTunnelType() ) );

       // Leonard's categories ....

       if ( this.isSpaceBlock() == true )
          buf.append( String.format("spaceblock = %s ... \n", this.getSpaceBlockType() ) );
       if ( this.isLineBlock() == true )
          buf.append( String.format(" lineblock = %s ... \n", this.getLineBlockType() ) );
       if ( this.isTrafficIntersection() == true )
          buf.append( String.format(" traffic intersection = %s ... \n", this.getTrafficIntersectionType() ) );

       // Summary categories ....

       if( nodes != null )
          buf.append( String.format("  no nodes = %d ... \n", nodes.size() ) );

       return buf.toString();
    }
}
