/*
 * ======================================================================
 * LibraryDimension.java: Assemble dimensions into a library format.
 *
 * Two types of mapping are supported:
 *
 * 1. A string-to-dimensions maps input strings to dimensions objects.
 * 2. A dimensions-to-string maps dimensions objects (and their attribute
 *    values) to names.
 *
 * The mappings are not symmetric: for example, "cm^2" is in the 
 * dimensions-to-string map but not the string-to-dimensions. The latter
 * is computed through the input grammar.
 *
 * Written by: Mark Austin
 * ======================================================================
 */

package whistle;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

public class LibraryDimension {
   public DimensionCheck unitsFlag;

   // Create singleton for string-to-dimension mapping ...

   public static HashMap<String,Dimension> s2d;

   // Create singleton for dimension-to-string mapping ...

   public static HashMap<Dimension,String> d2s;

   // Constructor methods ...

   public LibraryDimension() {
      this.s2d  = new HashMap<String,Dimension>();
      this.d2s  = new HashMap<Dimension,String>();
      unitsFlag = DimensionCheck.ON;
      buildLibrary();
   }

   // Retrieve library hashmaps ...

   public static HashMap getStringToLibraryDimen() {
      return s2d;
   }

   public static HashMap getDimenToStringLibrary() {
      return d2s;
   }

   // ===============================================================
   // Build library of well-known unit names and their properties ...
   // ===============================================================

   public void buildLibrary() {
      Dimension d01;
      String s01;

      // ---------------------------------------
      // Unit Constant ....
      // ---------------------------------------

      s01 = "";
      d01 = new Dimension ( 1e-0,  0,  0,  0, 0, 0, DimensionType.SI);
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Length units ...
      // ---------------------------------------

      s01 = "micron";
      d01 = new Dimension ( 1e-6,  1,  0,  0, 0, 0, DimensionType.SI);
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mm";
      d01 = new Dimension ( 1e-3,  1,  0,  0, 0, 0, DimensionType.SI);
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "cm";
      d01 = new Dimension ( 1e-2,  1,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "dm";
      d01 = new Dimension ( 1e-1,  1,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 =  "m";
      d01 = new Dimension ( 1e-0,  1,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "km";
      d01 = new Dimension ( 1e+3,  1,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Area units ...
      // ---------------------------------------

      s01 = "m^2";
      d01 = new Dimension ( 1.0e-0,  2,  0,  0, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );

      s01 = "cm^2";
      d01 = new Dimension ( 1.0e-4,  2,  0,  0, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );

      s01 = "mm^2";
      d01 = new Dimension ( 1.0e-6,  2,  0,  0, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Volume units ...
      // ---------------------------------------

      s01 = "m^3";
      d01 = new Dimension ( 1.0e-0,  3,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "cm^3";
      d01 = new Dimension ( 1.0e-6,  3,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mm^3";
      d01 = new Dimension ( 1.0e-9,  3,  0,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Miscellaneous length units ...
      // ---------------------------------------

      s01 = "m^4";
      d01 = new Dimension ( 1.0e-0,  4,  0,  0, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "cm^4";
      d01 = new Dimension ( 1.0e-8,  4,  0,  0, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // Velocity units ....
      // ---------------------------------------

      s01 = "m/sec";
      d01 = new Dimension ( 1.0e-0,  1,  0, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "cm/sec";
      d01 = new Dimension ( 1.0e-2,  1,  0, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "mm/sec";
      d01 = new Dimension ( 1.0e-3,  1,  0, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // Acceleration units ....
      // ---------------------------------------

      s01 = "m/sec^2";
      d01 = new Dimension ( 1.0e-0,  1,  0, -2, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "cm/sec^2";
      d01 = new Dimension ( 1.0e-2,  1,  0, -2, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "mm/sec^2";
      d01 = new Dimension ( 1.0e-3,  1,  0, -2, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // Viscosity units ...
      // ---------------------------------------

      s01 = "Pa.s";
      d01 = new Dimension ( 1.0e-0, -1,  1, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "kPa.s";
      d01 = new Dimension ( 1.0e-3, -1,  1, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // Miscellaneous Viscosity units ...
      // ---------------------------------------

      s01 = "1/Pa.s";
      d01 = new Dimension ( 1.0e-0,  1, -1,  1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "m/Pa.s";
      d01 = new Dimension ( 1.0e-0,  2, -1,  1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "cm/Pa.s";
      d01 = new Dimension ( 1.0e-2,  2, -1,  1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "mm/Pa.s";
      d01 = new Dimension ( 1.0e-3,  2, -1,  1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "m^2/Pa.s";
      d01 = new Dimension ( 1.0e-0,  3, -1,  1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "cm^2/Pa.s";
      d01 = new Dimension ( 1.0e-4,  3, -1,  1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // Flow-rate units ...
      // ---------------------------------------

      s01 = "m^3/sec";
      d01 = new Dimension ( 1.0e-0,  3,  0, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      s01 = "cm^3/sec";
      d01 = new Dimension ( 1.0e-6,  3,  0, -1, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // SI Mass units ...
      // ---------------------------------------

      s01 = "gram";
      d01 = new Dimension ( 1E-3,  0,  1,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kg";
      d01 = new Dimension (  1.0,  0,  1,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "Mg";
      d01 = new Dimension ( 1E+6,  0,  1,  0, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Units associated with densities ....
      // ---------------------------------------

      s01 = "kg/m^3";
      d01 = new Dimension ( 1e-0, -3,  1,  0, 0, 0, DimensionType.SI );
      d2s.put ( d01, s01 );
      s2d.put ( s01, d01 );

      // ---------------------------------------
      // SI_US Time units ...
      // ---------------------------------------

      s01 = "sec";
      d01 = new Dimension (    1.0,  0,  0,  1, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ms";
      d01 = new Dimension ( 0.0010,  0,  0,  1, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "min";
      d01 = new Dimension (   60.0,  0,  0,  1, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "hr";
      d01 = new Dimension ( 3600.0,  0,  0,  1, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Temperature units ...
      // ---------------------------------------

      s01 = "deg_C";
      d01 = new Dimension ( 1.0,  0,  0,  0, 1, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "DEG_C";
      d01 = new Dimension ( 1.0,  0,  0,  0, 1, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Frequency units ...
      // ---------------------------------------

      s01 = "Hz";
      d01 = new Dimension ( 1.0,  0,  0, -1, 0, 0, DimensionType.SI_US);
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI_US Speed units ...
      // ---------------------------------------

      s01 = "rpm";
      d01 = new Dimension ( 1.0/60.0,  0,  0, -1, 0, 0, DimensionType.SI_US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "cps";
      d01 = new Dimension (      1.0,  0,  0, -1, 0, 0, DimensionType.SI_US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Force units ...
      // ---------------------------------------

      s01 = "N";
      d01 = new Dimension (  1.0,  1,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kN";
      d01 = new Dimension ( 1E+3,  1,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kgf";
      d01 = new Dimension ( 9.80665,  1,  1, -2, 0, 0, DimensionType.SI ); 
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Pressure units ...
      // ---------------------------------------

      s01 = "Pa";
      d01 = new Dimension (  1.0, -1,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kPa";
      d01 = new Dimension ( 1E+3, -1,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "MPa";
      d01 = new Dimension ( 1E+6, -1,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "GPa";
      d01 = new Dimension ( 1E+9, -1,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Miscellaneous Pressure gradient units ...
      // ---------------------------------------

      s01 = "Pa/m";
      d01 = new Dimension (  1.0, -2,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "Pa/m/m";
      d01 = new Dimension (  1.0, -3,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Energy units ...
      // ---------------------------------------

      s01 = "Jou";
      d01 = new Dimension ( 1.0,  2,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kJ";
      d01 = new Dimension (1E+3,  2,  1, -2, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // SI Power units ....
      // ---------------------------------------

      s01 = "Watt";
      d01 = new Dimension (  1.0,  2,  1, -3, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kW";
      d01 = new Dimension ( 1e+3,  2,  1, -3, 0, 0, DimensionType.SI );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // US Length units ...
      // ---------------------------------------

      s01 = "in";
      d01 = new Dimension (  1.0,  1,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mil";
      d01 = new Dimension ( 1E-3,  1,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "micro_in";
      d01 = new Dimension ( 1E-6,  1,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ft";
      d01 = new Dimension ( 12.0,  1,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "yard";
      d01 = new Dimension ( 36.0,  1,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mile";
      d01 = new Dimension ( 63360.0,  1,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Area units ....
      // ---------------------------------------

      s01 = "in^2";
      d01 = new Dimension (       1.0,  2,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ft^2";
      d01 = new Dimension ( 12.0*12.0,  2,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "yard^2";
      d01 = new Dimension ( 36.0*36.0,  2,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Volume units ....
      // ---------------------------------------

      s01 = "in^3";
      d01 = new Dimension (            1.0,  3,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ft^3";
      d01 = new Dimension ( 12.0*12.0*12.0,  3,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "yard^3";
      d01 = new Dimension ( 36.0*36.0*36.0,  3,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Velocity units ....
      // ---------------------------------------

      s01 = "in/sec";
      d01 = new Dimension (  1.0,  1,  0, -1, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mil/sec";
      d01 = new Dimension ( 1E-3,  1,  0, -1, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "micro_in/sec";
      d01 = new Dimension ( 1E-6,  1,  0, -1, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ft/sec";
      d01 = new Dimension ( 12.0,  1,  0, -1, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "yard/sec";
      d01 = new Dimension ( 36.0,  1,  0, -1, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mile/sec";
      d01 = new Dimension ( 63360.0,  1,  0,  -1, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Velocity units ....
      // ---------------------------------------

      s01 = "in/sec^2";
      d01 = new Dimension (  1.0,  1,  0, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mil/sec^2";
      d01 = new Dimension ( 1E-3,  1,  0, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "micro_in/sec^2";
      d01 = new Dimension ( 1E-6,  1,  0, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ft/sec^2";
      d01 = new Dimension ( 12.0,  1,  0, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "yard/sec^2";
      d01 = new Dimension ( 36.0,  1,  0, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "mile/sec^2";
      d01 = new Dimension ( 63360.0,  1,  0,  -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // US Volume units ...
      // ---------------------------------------

      s01 = "gallon";
      d01 = new Dimension ( 3.785412E-3, 3,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "barrel";
      d01 = new Dimension ( 0.1589873,   3,  0,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // US Mass units ...
      // ---------------------------------------

      s01 = "lb";
      d01 = new Dimension ( 1.0,  0,  1,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "grain";
      d01 = new Dimension ( 1.0/7E+3,  0,  1,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ton";
      d01 = new Dimension ( 2E+3,  0,  1,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "klb";
      d01 = new Dimension ( 1E+3,  0,  1,  0, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // US Temperature units ...
      // ---------------------------------------

      s01 = "deg_F";
      d01 = new Dimension ( 1.0,  0,  0,  0, 1, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "DEG_F";
      d01 = new Dimension ( 1.0,  0,  0,  0, 1, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // US Force units ...
      // ---------------------------------------

      s01 = "lbf";
      d01 = new Dimension ( 1e+0,  1,  1, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "kips";
      d01 = new Dimension ( 1e+3,  1,  1, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // US Pressure units ...
      // ---------------------------------------

      s01 = "psi";
      d01 = new Dimension (  1.0, -1,  1, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "ksi";
      d01 = new Dimension ( 1e+3, -1,  1, -2, 0, 0, DimensionType.US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      // ---------------------------------------
      // Plane angle....
      // ---------------------------------------

      s01 = "deg";
      d01 = new Dimension ( 0.01745329,  0,  0,  0, 0, 1, DimensionType.SI_US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

      s01 = "rad";
      d01 = new Dimension (        1.0,  0,  0,  0, 0, 1, DimensionType.SI_US );
      s2d.put ( s01, d01 );
      d2s.put ( d01, s01 );

   }

   // ===========================================================
   // Utility functions for units operations
   //
   // -- setUnitsOn() : Set flag for units "on"
   // -- SetUnitsOff()    : Set flag for units "off"
   // -- CheckUnits()     : Flag to see if units are "on" or "off"
   // -- CheckUnitsType() : Check type of units
   //
   // ===========================================================

   public void setUnitsOn(){
      unitsFlag = DimensionCheck.ON;
   }

   public void setUnitsOff(){
      unitsFlag = DimensionCheck.OFF;
   }

   public DimensionCheck getUnits() {
      return unitsFlag;
   }

   // =========================================================
   // Exercise methods in LibraryDimension ....
   // =========================================================

   public static void main( String args [] ) {

      // Part 01: Create and print a simple dimension ....

      System.out.println("");
      System.out.println("*** Exercise methods in whistle/LibraryDimension ...");
      System.out.println("*** ============================================= ...");

      LibraryDimension lib = new LibraryDimension ();

      System.out.println("");
      System.out.println("*** Print string-to-dimensions mapping ...");
      System.out.println("");

      System.out.println( lib.getStringToLibraryDimen().toString() );

      System.out.println("");
      System.out.println("*** Print dimensions-to-string mapping ...");
      System.out.println("");

      System.out.println( lib.getDimenToStringLibrary().toString() );

      System.out.println("");
      System.out.println("*** Check round-tripping of names and dimensions ... ");
      System.out.println("");

      Dimension d01 = (Dimension) lib.getStringToLibraryDimen().get("mm");
      String s01     = (String)     lib.getDimenToStringLibrary().get(d01);
      System.out.printf("*** Searching for \"mm\"   --> Found %s\n", s01 );

      Dimension d02 = (Dimension) lib.getStringToLibraryDimen().get("kPa");
      String s02     = (String)     lib.getDimenToStringLibrary().get(d02);
      System.out.printf("*** Searching for \"kPa\"  --> Found %s\n", s02 );

      Dimension d03 = (Dimension) lib.getStringToLibraryDimen().get("cm");
      String s03     = (String)     lib.getDimenToStringLibrary().get(d03);
      System.out.printf("*** Searching for \"cm\" --> Found %s\n", s03 );

      System.out.println("");
      System.out.println("*** Create quantity and check library lookup ... ");
      System.out.println("");

      d01 = (Dimension) lib.getStringToLibraryDimen().get("mm");
      d02 = (Dimension) lib.getStringToLibraryDimen().get("cm");

      System.out.println ( "*** Dimension: " + d02.toString() ); 

      // Create working quantity objects .....

      Quantity q01 = new Quantity("q01");
      q01.setValue ( 10.0 );
      q01.setDimension( d02 );
      q01.setUnitsName( "cm" );

      System.out.println ( q01.toString() ); 

      Quantity q02 = new Quantity("q02");
      q02.setValue ( 20.0 );
      q02.setDimension( d02 );
      q02.setUnitsName( "cm" );

      System.out.println ( q02.toString() ); 

      Quantity q03 = q01.add( q02 );
      q03.setName("q03");
      Quantity q04 = q02.add( q01 );
      q04.setName("q04");

      Quantity q05 = q02.mult( q01 );
      q05.setName("q05");

      System.out.println ("*** q01.add(q02)  : " + q03.toString() ); 
      System.out.println ("*** q02.add(q01)  : " + q04.toString() ); 
      System.out.println ("*** q02.mult(q01) : " + q05.toString() ); 

      System.out.println("");
      System.out.println("*** ============================================= ...");
      System.out.println("*** Finished! ...");
   }
}

